//***************************************************************************//
//                                                                           //
// Copyright (C) 2001, International Business Machines Corporation.          //
//                                                                           //
// Author:  James Antognini, IBM Research, antognini@us.ibm.com.             //
//                                                                           //
// Filter driver.                                                            //
//                                                                           //
// This has been tested when set as an upper filter driver for a CD-ROM      //
// by d:\Win2KDDK\src\storage\filters\addfilter\addfilter.c, eg:             //
//                                                                           //
//   addfilter /device \Ide\IdeDeviceP0T1L0-a /add JAf2Driver                //
//                                                                           //
// In such a case, the UpperFilters value is set in                          //
// HKLM\System\CurrentControlSet\Enum\IDE\... (if it's an IDE drive).        //
//                                                                           //
// This has also been tested as an upper filter driver for a for a CD-ROM    //
// when loaded by LoadFilter (adapted from ctrl2cap from Sysinternals);      //
// in that case, all CDs under HKLM\SYSTEM\CurrentControlSet\Control\Class\  //
// {4D36E965-E325-11CE-BFC1-08002BE10318} have this driver                   //
// attached.                                                                 //
//                                                                           //
//                                                                           //
// Change History                                                            //
//                                                                           //
//   date         reason                                                     //
//   ----------   ---------------------------------------------------------- //
//                                                                           //
//   18 May 2001  Return correct status in IOCompletion routine; add a com-  //
//                pletion routine that shows how to signal.                  //
//                                                                           //
//***************************************************************************//
#include "wdm.h"
#define  JAf2DriverKernelMode 1                       // ensure MyDevExt obtained
#include "JAf2Driver.h"

//======================================================================
//                           D E F I N E S
//======================================================================-
// print macro that only turns on when debugging is on
#if DBG
#define DbgPrint(arg) DbgPrint arg
#else
#define DbgPrint(arg)
#endif

#define JAf2DriverVersion "2.00"

//======================================================================
//                          E X T E R N S
//======================================================================

//======================================================================
//                          F O R W A R D S
//======================================================================
NTSTATUS
JAf2DriverDispatch(
                   IN PDEVICE_OBJECT,
                   IN PIRP
                  );

NTSTATUS
JAf2DriverAddDevice(
                    IN PDRIVER_OBJECT,
                    IN PDEVICE_OBJECT         
                   );
                    
NTSTATUS
JAf2DriverDispatchPower(
                        IN PDEVICE_OBJECT,
                        IN PIRP          
                       );
                        
NTSTATUS
JAf2DriverDispatchPnP(
                      IN PDEVICE_OBJECT,
                      IN PIRP          
                     );
                        
NTSTATUS
JAf2DriverDispatchAny(
                      IN PDEVICE_OBJECT,
                      IN PIRP
                     );

NTSTATUS
JAf2DriverDispatchWMI(
                      IN PDEVICE_OBJECT,
                      IN PIRP
                     );

NTSTATUS
JAf2DriverCompletionRtn(
                        IN PDEVICE_OBJECT,
                        IN PIRP,
                        IN PVOID
                       );

NTSTATUS
JAf2DriverCompletionRtn2(
                         IN PDEVICE_OBJECT,
                         IN PIRP,
                         IN PVOID
                        );

VOID
JAf2DriverUnload(IN PDRIVER_OBJECT);

int
JAf2DriverTechnique(
                    IN PDEVICE_OBJECT,
                    IN PIRP,
                    IN PVOID,
                    IN ULONG,
                    IN PVOID,
                    IN ULONG
                   );

//***************************************************************************//
//                                                                           //
// Globals.                                                                  //
//                                                                           //
//***************************************************************************//

//***************************************************************************//
//                                                                           //
// Globals for testing.                                                      //
//                                                                           //
//***************************************************************************//

PCHAR pJAf2DriverInfo =                               // Time and date of compilation.
        &"JAf2Driver v" JAf2DriverVersion ", compiled " __TIME__ " " __DATE__;

PDRIVER_OBJECT   pMstrJAf2DrvObj = NULL;

PDEVICE_OBJECT   pJAf2MstrDevObj = NULL;              // "Master" device object, created in DriverEntry.

PVOID            pZitsX = (PVOID)0x12345678;          // testing.

// #pragma warning(push)
// #pragma warning(disable : 4101)                       // suppress unreference local variable informational warning.

//----------------------------------------------------------------------
//
// DriverEntry
//
//----------------------------------------------------------------------

NTSTATUS
DriverEntry(
            IN PDRIVER_OBJECT  pDriverObject,
            IN PUNICODE_STRING pRegistryPath
           )
{
 NTSTATUS                        status = STATUS_SUCCESS;
 UNICODE_STRING                  DeviceNameUniString,
                                 DeviceLinkUniString,
                                 DriverName,
                                 WkUniStr;
 PDEVICE_OBJECT                  pDeviceObject = NULL;
 pJAf2DeviceExtension            pDevExt;
 ULONG                           i;
 BOOLEAN                         bHaveSymLink = FALSE;
 WCHAR                     const backslash = L'\\';   // delimiter in Unicode.          
 WCHAR                     const constDeviceName[] = L"\\Device\\",                 
                                 constDeviceLink[] = L"\\DosDevices\\";   
 PWSTR                           pIdx;
 ANSI_STRING                     LclDrvName;
  
 DbgPrint((JAf2DrvNm " v" JAf2DriverVersion ":  Entered\n"));

 // Invoke an attached debugger.  If there's none, continue.

 _try                                                 // per Mark Roddy, comp.os.ms-windows.programmer.nt.kernel-mode, 27 Apr 2001.
    {                                     
     DbgBreakPoint();                     
    }                                     
     _except(EXCEPTION_EXECUTE_HANDLER)   
      {                                   
      }                                   

 RtlZeroMemory(&LclDrvName, sizeof(LclDrvName));      // clear ansi string descriptor, ensure safe exit.

 DeviceNameUniString.Buffer = NULL;                   // ensure safe exit by branch to label 'done'.
 DeviceLinkUniString.Buffer = NULL;
 LclDrvName.Buffer = NULL;
  
 // Is the next important?
  
 DriverName.Buffer = NULL;

 // Having a device object attached to the driver object when the CDROM is disabled prevents the driver from being
 // unloaded.

 // Find the driver name in the registry path information.  The following assumes the string to be searched ends with a
 // backslash followed by 1 or more WCHARs that are the Unicode name.
  
 for (                                                // find last delimiter by working back from end.       
      pIdx =                                          // point to last WCHAR in buffer.                      
        pRegistryPath->Buffer + (pRegistryPath->Length / sizeof(WCHAR)) - 1;
      ;
      pIdx--                                          // decrement by WCHAR width.                           
     )                                                                                                       
    if (backslash==*pIdx)                             // is current WCHAR a delimiter?                       
      break;                                                                                                 

 // Build Unicode descriptor of driver name.
  
 pIdx++;                                              // point to just after backslash.
  
 DriverName.Length =                                  // calculate length.              
   pRegistryPath->Length -                                     
   ((pIdx - pRegistryPath->Buffer) * sizeof(WCHAR)); 
  
 DriverName.MaximumLength = DriverName.Length +       // figure in length of terminator.
                            sizeof(WCHAR);
 DriverName.Buffer = pIdx;                            // point to driver name.

 status =                                             // get driver name in ansi.
   RtlUnicodeStringToAnsiString(
                                &LclDrvName,          // descriptor.
                                &DriverName,          // Unicode string to convert.
                                TRUE                  // allocate space.
                               );

 if (!NT_SUCCESS(status))                             // a problem?
   {
    DbgPrint((JAf2DrvNm " DriverEntry:  Failed to convert driver name!\n"));
    goto done;
   }

 DeviceNameUniString.Length = 0;                      // show current length of device name.

 DeviceNameUniString.MaximumLength =                  // show maximum length of device name.
   sizeof(constDeviceName) -
   sizeof(WCHAR)           +                          // don't need size of first terminator (that is, terminator of constDeviceName).
   DriverName.MaximumLength;

 DeviceNameUniString.Buffer =                         // get space for device name and for ASCII driver name.
   ExAllocatePoolWithTag(           
                         PagedPool,
                         DeviceNameUniString.MaximumLength,
                         'f2JA'
                        );

 if (NULL==DeviceNameUniString.Buffer)                // failed to get storage? 
   {
    DbgPrint(("%s DriverEntry:  Failed to get storage for device name!\n", LclDrvName.Buffer));
    goto done;
   }

 // Now build proper device name, eg, '\Device\Driver'.

 RtlAppendUnicodeToString(                            // initialize by appending L"\\Device\\" to empty string.
                          &DeviceNameUniString,
                          constDeviceName
                         );

 RtlAppendUnicodeStringToString(                      // append driver name.
                                &DeviceNameUniString,
                                &DriverName
                               );
                                
 //
 // Set up the device
 //

 status = IoCreateDevice(
                         pDriverObject,
                         sizeof(JAf2DeviceExtension),
                         &DeviceNameUniString,
                         FILE_DEVICE_UNKNOWN,         // per Walt Oney, p 403.
                         0,
                         FALSE,                       // not exclusive.
                         &pDeviceObject
                        );

 if (FALSE==NT_SUCCESS(status))                       // a problem?
   {
    DbgPrint(("%s DriverEntry:  Failed to create the device!\n", LclDrvName.Buffer));
    goto done;
   }

 DbgPrint(("%s DriverEntry:  Main device at 0x%08x\n", LclDrvName.Buffer, pDeviceObject));

 pDevExt = (pJAf2DeviceExtension)pDeviceObject->DeviceExtension;

 if (NULL==pDevExt)
   {
    DbgPrint (("%s DriverEntry:  pDevExt = 0!\n", LclDrvName.Buffer));
    status = STATUS_UNSUCCESSFUL;

    goto done;
   }

 RtlZeroMemory(                                       // clear device extension.
               pDevExt,
               sizeof(JAf2DeviceExtension)
              );

 pDevExt->JAf2UniDeviceName.Length =                  // save length of Unicode driver name.
   DeviceNameUniString.Length;                                                            
 pDevExt->JAf2UniDeviceName.MaximumLength =           // save maximum length of Unicode driver name.
   DeviceNameUniString.MaximumLength;                                                            
 pDevExt->JAf2UniDeviceName.Buffer =                  // save address of Unicode driver name.
   DeviceNameUniString.Buffer;                                                                   
   
 memcpy(&pDevExt->DriverName,                         // copy ansi-string descriptor.
        &LclDrvName,
        sizeof(pDevExt->DriverName)
       );
   
 pDevExt->MyNumber = 1;
 pDevExt->MyArray[0] = 64;
 strcpy(pDevExt->MyStuff,"James Antognini");

 // Create a symbolic link to gain access to this driver/device, eg, '\DosDevices\Driver'.

 DeviceLinkUniString.Buffer =                         // get space for symbolic link name.
   ExAllocatePoolWithTag(           
                         PagedPool,
                         sizeof(constDeviceLink) +
                         DriverName.Length, 
                         'f2JA'
                        );

 if (NULL==DeviceLinkUniString.Buffer)                // failed to get storage? 
   {
    DbgPrint(("%s DriverEntry:  Failed to get storage for device link!\n", LclDrvName.Buffer));

    goto done;
   }

 DeviceLinkUniString.Length = 0;                      // show current length.
 DeviceLinkUniString.MaximumLength =                  // show maximum length.
   sizeof(constDeviceLink) -
   sizeof(WCHAR)           +                          // don't need size of first terminator (that is, terminator of constDeviceLink).
   DriverName.MaximumLength;

 RtlAppendUnicodeToString(                            // initialize by appending L"\\DosDevices\\" to empty string.
                          &DeviceLinkUniString,
                          constDeviceLink
                         );

 RtlAppendUnicodeStringToString(                      // append driver name.
                                &DeviceLinkUniString,
                                &DriverName
                               );
                                
 status = IoCreateSymbolicLink(
                               &DeviceLinkUniString,
                               &DeviceNameUniString
                              );

 if (FALSE==NT_SUCCESS(status))                       // a problem?
   {
    DbgPrint(("%s DriverEntry:  IoCreateSymbolicLink failed\n", LclDrvName.Buffer));

    goto done;
   }

 bHaveSymLink = TRUE;

 pDevExt->JAf2UniSymLinkName.Length =                 // save length of Unicode symbolic link name.
   DeviceLinkUniString.Length;                                                            
 pDevExt->JAf2UniSymLinkName.MaximumLength =          // save maximum length of Unicode symbolic link name.
   DeviceLinkUniString.MaximumLength;                                                            
 pDevExt->JAf2UniSymLinkName.Buffer =                 // save address of Unicode symbolic link name.
   DeviceLinkUniString.Buffer;                                                                   
   
 // Minimally support all types of IRP (taken from Walter Oney's "Programming the Microsoft Windows Driver Model,"
 // p 402.
   
 for (i = 0; i <= IRP_MJ_MAXIMUM_FUNCTION; i++)
   pDriverObject->MajorFunction[i] = JAf2DriverDispatchAny;

 //
 // Export dispatch points for all routines that must be supported by JAf2Driver.
 //

 pDriverObject->MajorFunction[IRP_MJ_CREATE]           =
   pDriverObject->MajorFunction[IRP_MJ_CLOSE]          =
   pDriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = JAf2DriverDispatch;

 pDriverObject->MajorFunction[IRP_MJ_PNP]              = JAf2DriverDispatchPnP;
 pDriverObject->MajorFunction[IRP_MJ_POWER]            = JAf2DriverDispatchPower;

 // Haven't got WMI working yet (8 May 2001).

// pDriverObject->MajorFunction[IRP_MJ_SYSTEM_CONTROL]   = JAf2DriverDispatchWMI;

 // Export entry points for other functions.

 pDriverObject->DriverExtension->AddDevice             = JAf2DriverAddDevice;
 pDriverObject->DriverUnload                           = JAf2DriverUnload;

 DbgPrint(("%s DriverEntry:  Initialization complete\n", LclDrvName.Buffer));

done:

 if (NT_SUCCESS(status))                              // no problem?
   {                                                
    pMstrJAf2DrvObj = pDriverObject;
    pJAf2MstrDevObj = pDeviceObject;
   }
 else
   {
    if (TRUE==bHaveSymLink)                           // symbolic link established?
      IoDeleteSymbolicLink(&DeviceLinkUniString);

    if (NULL!=DeviceLinkUniString.Buffer)             // free working storage for symbolic link name.
      ExFreePool(DeviceLinkUniString.Buffer);

    if (NULL!=DeviceNameUniString.Buffer)             // free working storage for device name.
      ExFreePool(DeviceNameUniString.Buffer);

    if (NULL!=LclDrvName.Buffer)                      // free working storage for ansi driver name.
      RtlFreeAnsiString(&LclDrvName);              

    if (NULL!=pDeviceObject)                          // was a device object created?
      IoDeleteDevice(pDeviceObject);
   }

 return status;
}

/**************************************************************************************************/
/*                                                                                                */
/**************************************************************************************************/
NTSTATUS
JAf2DriverAddDevice(
                    IN PDRIVER_OBJECT pDriverObject,
                    IN PDEVICE_OBJECT pPDO           
                   )
{
 NTSTATUS                        status = STATUS_SUCCESS;
 PDEVICE_OBJECT                  pDeviceObject = NULL,
                                 pLclNextLowerDevObj;
 pJAf2DeviceExtension            pDevExt,
                                 pMstrDevExt = (pJAf2DeviceExtension)pJAf2MstrDevObj->DeviceExtension;
 BOOLEAN                         bAttachedDevObj = FALSE;
                                  
 DbgPrint(("%s AddDevice:  Entered\n", pMstrDevExt->DriverName.Buffer));
                                  
 status = IoCreateDevice(
                         pDriverObject,
                         sizeof(JAf2DeviceExtension),
                         NULL,                        // no name.
                         FILE_DEVICE_CD_ROM,                                    
                         FILE_AUTOGENERATED_DEVICE_NAME,  // OS will produce a name.
                         FALSE,                       // not exclusive.
                         &pDeviceObject
                        );

 if (FALSE==NT_SUCCESS(status))                       // a problem?
   {
    DbgPrint(("%s AddDevice:  Failed to create the device!\n", pMstrDevExt->DriverName.Buffer));
    goto done;
   }

 pDevExt = (pJAf2DeviceExtension)pDeviceObject->DeviceExtension;

 if (NULL==pDevExt)
   {
    DbgPrint (("%s AddDevice:  pDevExt = 0!\n", pMstrDevExt->DriverName.Buffer));
    status = STATUS_UNSUCCESSFUL;
    goto done;
   }

 DbgPrint(("%s AddDevice:  Added device at 0x%08x\n", pMstrDevExt->DriverName.Buffer, pDeviceObject));

 RtlZeroMemory(                                       // clear device extension.
               pDevExt,
               sizeof(JAf2DeviceExtension)
              );

 pDevExt->pMstrDevObj = pJAf2MstrDevObj;              // save address of master device object.

 IoInitializeRemoveLock(                              // initialize removal lock.
                        &pDevExt->RemoveLock,
                        (ULONG)'kLAJ',                // identifying tag.
                        1,                            // should hold note more than 1 minute.
                        5                             // in checked build, max number outstanding requests.
                       );

 pDevExt->pPhysDevObj = pPDO;                         // point to physical device object.

 pLclNextLowerDevObj =                                // attach to current highest object (physical device object).
   pDevExt->pNextLowerDevObj =                        //   needed for IoDetachDevice().
     IoAttachDeviceToDeviceStack(
                                 pDeviceObject,
                                 pPDO
                                );
                                                                                                                     
 if (NULL==pLclNextLowerDevObj)                       // a problem?
   {                                                                                                                 
    DbgPrint(("%s AddDevice:  IoAttachDeviceToDeviceStack failed, rc = 0x%08x\n", pMstrDevExt->DriverName.Buffer, status));
                                                                                                                     
    status = STATUS_UNSUCCESSFUL;
                                                                                                                     
    goto done;                                                                                                       
   }                                                                                                                 
                                                                                                                     
 bAttachedDevObj = TRUE;                                                                                             

 pDeviceObject->Flags &= ~(                           // ensure 4 bits off in filter driver's device object.   
                           DO_DIRECT_IO     |                                                                  
                           DO_BUFFERED_IO   |                                                                  
                           DO_POWER_PAGABLE |                                                                  
                           DO_POWER_INRUSH                                                                     
                          );                                                                                   
                                                                                                               
 pDeviceObject->Flags |= pLclNextLowerDevObj->Flags & // per Oney's errata on http://www.oneysoft.com/wdmerr.htm.
                         (                                                                                     
                          DO_DIRECT_IO     |                                                                   
                          DO_BUFFERED_IO                                                                     
                         );                                                                                    
                                                                                                               
 pDeviceObject->Flags |= DO_POWER_PAGABLE;            // per Oney's errata on http://www.oneysoft.com/wdmerr.htm.
                                                                                                               
 pDeviceObject->DeviceType =                          // per Oney, p 403.
   pLclNextLowerDevObj->DeviceType;                          
                                                                                                               
 pDeviceObject->Characteristics =                     // per Oney, p 403.                                      
   pLclNextLowerDevObj->Characteristics;
                                                                                                               
 // Haven't got WMI working yet (8 May 2001).
                                                                                                               
 // status = IoWMIRegistrationControl(                // per Oney, p 456.                                         
 //                                   pDeviceObject,                                                              
 //                                   WMIREG_ACTION_REGISTER                                                      
 //                                  );                                                                           

 // if (!NT_SUCCESS(status))
 //   {
 //    DbgPrint((JAf2DrvNm " AddDevice: WMI reg failed, status = 0x%08x\n", status));
 //    goto done;
 //   }
                                                                                                               
 pDeviceObject->Flags &= ~DO_DEVICE_INITIALIZING;     // show device is initialized.                           

done:

 if (NT_SUCCESS(status))                              // no problem?
   {                                                 
//  DbgPrint((JAf2DrvNm " AddDevice: ulOpen        at 0x%08x\n", &pDevExt->ulOpen));
//  DbgPrint((JAf2DrvNm " AddDevice: ulClose       at 0x%08x\n", &pDevExt->ulClose));
//  DbgPrint((JAf2DrvNm " AddDevice: ulOpenFailed  at 0x%08x\n", &pDevExt->ulOpenFailed));
//  DbgPrint((JAf2DrvNm " AddDevice: ulCloseFailed at 0x%08x\n", &pDevExt->ulCloseFailed));

    InterlockedIncrement(                             // bump added device count in master device object extension.
                         &((pJAf2DeviceExtension)pJAf2MstrDevObj->DeviceExtension)->ulAddedDevObj
                        );
   }
 else
   {
    if (TRUE==bAttachedDevObj)                        // attached to next-lower device object?      
      IoDetachDevice(pLclNextLowerDevObj);                                                            
       
    if (NULL!=pDeviceObject)                          // was a device object created?
      IoDeleteDevice(pDeviceObject);
   }

 DbgPrint(("%s AddDevice:  Done, final status = 0x%08x\n", pMstrDevExt->DriverName.Buffer, status));

 return status;        
}
 
/**************************************************************************************************/
/*                                                                                                */
/* This is called for Open, Close and DeviceIoControl.                                            */
/*                                                                                                */
/**************************************************************************************************/
NTSTATUS
JAf2DriverDispatch(
                   IN PDEVICE_OBJECT pDeviceObject,
                   IN PIRP           pIrp
                  )
{
 NTSTATUS                status = STATUS_SUCCESS;
 PIO_STACK_LOCATION      pIrpStack;
 PVOID                   pInBuffer,
                         pOutBuffer,
                         ptr;
 ULONG                   lenInBuffer,
                         lenOutBuffer,
                         IoControlCode,
                         lclRc;
 pJAf2DeviceExtension    pDevExt =                    // address of device object extension.
                           (pJAf2DeviceExtension)pDeviceObject->DeviceExtension,
                         pMstrDevExt = (pJAf2DeviceExtension)pDevExt->pMstrDevObj->DeviceExtension;
 PLONG                   pZits;
 KIRQL                   SaveIrql;
 BOOLEAN                 bCompleteReq = FALSE,
                         bUseCompletionRtn = FALSE,
                         bHaveRemoveLock = FALSE;
                          
 // Get a pointer to the current location in the Irp.  This is where the function codes and 
 // parameters are located.

 pIrpStack = IoGetCurrentIrpStackLocation(pIrp);

 // Get the pointer to the input/output buffer and its length

 pInBuffer          = pIrp->AssociatedIrp.SystemBuffer;
 lenInBuffer        = pIrpStack->Parameters.DeviceIoControl.InputBufferLength;
 pOutBuffer         = pIrp->AssociatedIrp.SystemBuffer;
 lenOutBuffer       = pIrpStack->Parameters.DeviceIoControl.OutputBufferLength;
 IoControlCode      = pIrpStack->Parameters.DeviceIoControl.IoControlCode;

// DbgPrint((JAf2DrvNm " Dispatch:       MajorFunction = 0x%02x\n", pIrpStack->MajorFunction));

 status = IoAcquireRemoveLock(
                              &pDevExt->RemoveLock,
                              NULL                    // don't supply caller's tag.                              
                             );

 if (!NT_SUCCESS(status))                             // no good?
   {
    DbgPrint(("%s Dispatch:  Failed to get RemoveLock!\n", pMstrDevExt->DriverName.Buffer));

    pIrp->IoStatus.Status = status;              
    bCompleteReq = TRUE;                              // so request to be completed.
       
    goto done; 
   }

 bHaveRemoveLock = TRUE;

 switch(pIrpStack->MajorFunction)
   {
    case IRP_MJ_DEVICE_CONTROL:                       // DeviceIoControl.

//    DbgPrint((JAf2DrvNm " Dispatch:       Dev IoControlCode = 0x%08x\n", IoControlCode));

      switch(IoControlCode)
        {
         default:
           ;
        }                                             // end switch(IoControlCode)

      break;

    default:
      ;
   }

done:

 if (NT_SUCCESS(status))                              // went OK?
   {
    if (IRP_MJ_CREATE==pIrpStack->MajorFunction)      // Open?
      InterlockedIncrement(&pDevExt->ulOpen);         // bump Open count.
    else
      if (IRP_MJ_CLOSE==pIrpStack->MajorFunction)     // Close?
        InterlockedIncrement(&pDevExt->ulClose);      // bump Close count.
   }
 else
   {
    if (IRP_MJ_CREATE==pIrpStack->MajorFunction)      // Open?
      InterlockedIncrement(&pDevExt->ulOpenFailed);   // bump failed Open count.
    else
      if (IRP_MJ_CLOSE==pIrpStack->MajorFunction)     // Close?
        InterlockedIncrement(&pDevExt->ulCloseFailed); // bump failed Close count.
   }
        
 if (TRUE==bCompleteReq)                              // I/O request to be completed?
   {
    IoCompleteRequest(pIrp, IO_NO_INCREMENT);         // complete request.
   }
 else
   {
//  bUseCompletionRtn = TRUE;                         // control whether IoCompletion routine is run.
     
    if (TRUE==bUseCompletionRtn)                      // have a completion routine? 
      {                                               
       _asm nop                                       // for debugging, 'cause 2 macros next don't work well with WinDbg. 

       IoCopyCurrentIrpStackLocationToNext(pIrp);     // copy parameters from current stack position to next-lower position.
                                                                                       
       IoSetCompletionRoutine(                                                         
                              pIrp,                                                    
                              (PIO_COMPLETION_ROUTINE)JAf2DriverCompletionRtn,         
                              pZitsX,                 // testing context.              
                              TRUE,                   // invoke on success.            
                              TRUE,                   // invoke on error.              
                              TRUE                    // invoke on cancel.             
                             );                                                        
      }
    else
      {                                               
       IoSkipCurrentIrpStackLocation(pIrp);           // copy parameters from current stack position to next-lower position.
                                                      //   No completion routine.
      }
     
    status =     
      IoCallDriver(                                   // call next-lower driver.
                   pDevExt->pNextLowerDevObj,
                   pIrp
                  );
     
   }

 if (TRUE==bHaveRemoveLock)                           // have removal lock?
   IoReleaseRemoveLock(
                       &pDevExt->RemoveLock,
                       NULL                           // didn't supply caller's tag.
                      );

// DbgPrint(("%s Dispatch:       Final status = 0x%08x\n", pMstrDevExt->DriverName.Buffer, status));
// DbgPrint(("                           Irp status   = 0x%08x\n", pIrp->IoStatus.Status));

 return status;
}

/**************************************************************************************************/
/*                                                                                                */
/**************************************************************************************************/
NTSTATUS
JAf2DriverDispatchPower(
                        IN PDEVICE_OBJECT pDeviceObject,
                        IN PIRP           pIrp
                       )
{
 NTSTATUS                status = STATUS_SUCCESS;
 PIO_STACK_LOCATION      pIrpStack;
 pJAf2DeviceExtension    pDevExt =                    // address of device object extension.
                           (pJAf2DeviceExtension)pDeviceObject->DeviceExtension,
                         pMstrDevExt = (pJAf2DeviceExtension)pDevExt->pMstrDevObj->DeviceExtension;
 BOOLEAN                 bCompleteReq = FALSE,
                         bUseCompletionRtn = FALSE,
                         bHaveRemoveLock = FALSE;
  
 pIrpStack = IoGetCurrentIrpStackLocation(pIrp);

// DbgPrint(("%s DispatchPower:  MinorFunction = 0x%02x\n", pMstrDevExt->DriverName.Buffer, pIrpStack->MinorFunction));

 PoStartNextPowerIrp(pIrp);                           // before IoAcquireRemoveLock, per Oney, p 405.

 status = IoAcquireRemoveLock(
                              &pDevExt->RemoveLock,
                              NULL                    // don't supply caller's tag.                              
                             );

 if (!NT_SUCCESS(status))                             // no good?
   {
    DbgPrint(("%s DispatchPower:  Failed to get RemoveLock!\n", pMstrDevExt->DriverName.Buffer));

    pIrp->IoStatus.Status = status;              
    bCompleteReq = TRUE;                              // so request to be completed.
       
    goto done; 
   }

 bHaveRemoveLock = TRUE;

 IoSkipCurrentIrpStackLocation(pIrp);                      

 status =
   PoCallDriver(                                      // call next-lower driver.
                pDevExt->pNextLowerDevObj,
                pIrp
               );

done:

 if (TRUE==bCompleteReq)                              // I/O request to be completed?
   {
    IoCompleteRequest(pIrp, IO_NO_INCREMENT);         // complete request.
   }

 if (TRUE==bHaveRemoveLock)                           // have removal lock?
   IoReleaseRemoveLock(
                       &pDevExt->RemoveLock,
                       NULL                           // didn't supply caller's tag.
                      );

// DbgPrint(("%s DispatchPower:  Final status = 0x%08x\n", pMstrDevExt->DriverName.Buffer, status));
// DbgPrint(("                           Irp status   = 0x%08x\n", pIrp->IoStatus.Status));

 return status;
}
 
/**************************************************************************************************/
/*                                                                                                */
/**************************************************************************************************/
NTSTATUS
JAf2DriverDispatchPnP(
                      IN PDEVICE_OBJECT pDeviceObject,
                      IN PIRP           pIrp
                     )
{
 NTSTATUS                status;
 PIO_STACK_LOCATION      pIrpStack;
 pJAf2DeviceExtension    pDevExt =                    // address of device object extension.
                           (pJAf2DeviceExtension)pDeviceObject->DeviceExtension,
                         pMstrDevExt = (pJAf2DeviceExtension)pDevExt->pMstrDevObj->DeviceExtension;
 BOOLEAN                 bHaveRemoveLock = FALSE,
                         bCompleteReq = FALSE;
 UCHAR                   saveMinFn;
  
 pIrpStack = IoGetCurrentIrpStackLocation(pIrp);

// DbgPrint(("%s DispatchPnP:    MinorFunction = 0x%02x\n", pMstrDevExt->DriverName.Buffer, pIrpStack->MinorFunction));

 status = IoAcquireRemoveLock(
                              &pDevExt->RemoveLock,
                              NULL                    // don't supply caller's tag.                              
                             );

 if (!NT_SUCCESS(status))                             // no good?
   {
    DbgPrint(("%s DispatchPnP:  Failed to get RemoveLock!\n", pMstrDevExt->DriverName.Buffer));

    pIrp->IoStatus.Status = status;              
    bCompleteReq = TRUE;                              // so request to be completed.
       
    goto done; 
   }

 bHaveRemoveLock = TRUE;

 saveMinFn = pIrpStack->MinorFunction;                // save information from stack, since it may be changed after
                                                      //   IoCallDriver.
 switch(saveMinFn)
   {
    // For next 2 cases and action, see \src\wdm\general\filter\pnp.c, routine VA_Pnp.
     
    case IRP_MN_QUERY_STOP_DEVICE:    
    case IRP_MN_QUERY_REMOVE_DEVICE:  
     
      pIrp->IoStatus.Status = STATUS_SUCCESS;         // test.  30.04.01.

      break;
    
    // For next case and action, see \src\wdm\general\filter\pnp.c, routine VA_Pnp.
    
    case IRP_MN_SURPRISE_REMOVAL:    
    
      pIrp->IoStatus.Status = STATUS_SUCCESS;         // test.  30.04.01.

      break;
    
    case IRP_MN_STOP_DEVICE:
     
      _asm nop
       
      break;
       
    case IRP_MN_REMOVE_DEVICE:                        // Next follows \src\general\toaster\func\toaster.c, 
                                                      //   routine ToasterDispatchPnp.     
//    DbgPrint(("%s DispatchPnP:  IRP_MN_REMOVE_DEVICE\n", pMstrDevExt->DriverName.Buffer));

      // Note:  Detach/delete will be done after calling next-lower driver.

      // Haven't got WMI working yet (8 May 2001).
                                                                                                               
//    status = IoWMIRegistrationControl(              
//                                      pDeviceObject,
//                                      WMIREG_ACTION_DEREGISTER
//                                     );
                                                                                                               
    default:
      ;
   }

 IoSkipCurrentIrpStackLocation(pIrp);                 // copy parameters from current stack position to next-lower position.
                                                      //   No completion routine.

 status =
   IoCallDriver(                                      // call next-lower driver.
                pDevExt->pNextLowerDevObj,
                pIrp
               );

// DbgPrint(("%s DispatchPnP:    Status from IoCallDriver = 0x%08x\n", pMstrDevExt->DriverName.Buffer, status));

 if (IRP_MN_REMOVE_DEVICE==saveMinFn)                 // remove device?
   {
    pJAf2DeviceExtension pMstrDevExt =                // address of device object extension.
                          (pJAf2DeviceExtension)pJAf2MstrDevObj->DeviceExtension;
                           
    IoReleaseRemoveLock(                              // per Oney, p 405.
                        &pDevExt->RemoveLock,
                        NULL                          // didn't supply caller's tag.
                       );

    bHaveRemoveLock = FALSE;

//  DbgPrint(("%s DispatchPnP:    Open count       = %08d, Close count       = %08d\n", pMstrDevExt->DriverName.Buffer, pDevExt->ulOpen, pDevExt->ulClose));
//  DbgPrint(("%s DispatchPnP:    FailedOpen count = %08d, FailedClose count = %08d\n", pMstrDevExt->DriverName.Buffer,
//            pDevExt->ulOpenFailed, pDevExt->ulCloseFailed));

    // Note: Putting detach/delete at end follows \src\input\kbfiltr\kbfiltr.c, in routine KbFilter_PnP,
    //       also in \src\storage\filters\diskperf\diskperf.c, in routine DiskPerfRemoveDevice.

    IoDetachDevice(pDevExt->pNextLowerDevObj);        // detach from next-lower device.
     
    IoDeleteDevice(pDeviceObject);                    // delete device created in AddDevice.

    InterlockedDecrement(                             // reduce added device count in master device object extension.
                         &pMstrDevExt->ulAddedDevObj 
                        );

    if (0==pMstrDevExt->ulAddedDevObj)                // no more added device objects?
      {                                               // clean up master device object.
       PDEVICE_OBJECT pMstrTempDevObj = pJAf2MstrDevObj;
        
       // Delete the symbolic link for master device object.

       IoDeleteSymbolicLink(&pMstrDevExt->JAf2UniSymLinkName);

       // Free storage for symbolic link name.

       ExFreePool(pMstrDevExt->JAf2UniSymLinkName.Buffer);

       // Free storage for master device object name.

       ExFreePool(pMstrDevExt->JAf2UniDeviceName.Buffer);

       // Free storage for ansi driver name.             

       RtlFreeAnsiString(&pMstrDevExt->DriverName);

       // Reset in-module pointer to master device object and delete master device object.

       pJAf2MstrDevObj = NULL;

       IoDeleteDevice(pMstrTempDevObj);
      }

//  DbgPrint((JAf2DrvNm " DispatchPnP:    Finished detach/delete\n"));
   }
    
done:    
    
 if (TRUE==bCompleteReq)                              // I/O request to be completed?
   {
    IoCompleteRequest(pIrp, IO_NO_INCREMENT);         // complete request.
   }

 if (TRUE==bHaveRemoveLock)                          // have removal lock?
   IoReleaseRemoveLock(
                       &pDevExt->RemoveLock,
                       NULL                           // didn't supply caller's tag.
                      );

// DbgPrint((JAf2DrvNm " DispatchPnP:    Final status = 0x%08x\n", status));
// DbgPrint(("                           Irp status   = 0x%08x\n", pIrp->IoStatus.Status));

 return status;
}
 
/**************************************************************************************************/
/*                                                                                                */
/* Some of this from FilterPass routine in filter.c in d:\Win2KDDK\src\general\toaster\filter\.   */
/*                                                                                                */
/* Note:  The effect of this routine is to make synchronous (with respect to caller of this       */
/*        dispatch routine) the Irp handling.                                                     */
/*                                                                                                */
/**************************************************************************************************/
NTSTATUS
JAf2DriverDispatchAny(
                      IN PDEVICE_OBJECT pDeviceObject,
                      IN PIRP           pIrp
                     )
{
 NTSTATUS                status = STATUS_SUCCESS;
 PIO_STACK_LOCATION      pIrpStack;
 BOOLEAN                 bCompleteReq = FALSE,
                         bUseCompletionRtn = FALSE,
                         bHaveRemoveLock = FALSE;
 pJAf2DeviceExtension    pDevExt =                    // address of device object extension.
                           (pJAf2DeviceExtension)pDeviceObject->DeviceExtension,
                         pMstrDevExt = (pJAf2DeviceExtension)pDevExt->pMstrDevObj->DeviceExtension;
 KEVENT                  lclEvent;

 pIrpStack = IoGetCurrentIrpStackLocation(pIrp);

 DbgPrint(("%s DispatchAny:    MajorFunction = 0x%02x, MinorFunction = 0x%02x\n", pMstrDevExt->DriverName.Buffer,
           pIrpStack->MajorFunction, pIrpStack->MinorFunction));

 status = IoAcquireRemoveLock(
                              &pDevExt->RemoveLock,
                              NULL                    // don't supply caller's tag.                              
                             );

 if (!NT_SUCCESS(status))                             // no good?
   {
    DbgPrint(("%s DispatchAny:  Failed to get RemoveLock!\n", pMstrDevExt->DriverName.Buffer));

    pIrp->IoStatus.Status = status;              
    bCompleteReq = TRUE;                              // so request to be completed.
       
    goto done; 
   }

 bHaveRemoveLock = TRUE;

 switch(pIrpStack->MajorFunction)
   {
    default:
      ;
   }

 if (TRUE==bCompleteReq)                              // I/O request to be completed?
   {
    IoCompleteRequest(pIrp, IO_NO_INCREMENT);         // complete request.
    bCompleteReq = FALSE;                             // don't complete a second time.
   }
 else
   {
    bUseCompletionRtn = TRUE;                         // control whether IoCompletion routine is run.

    if (TRUE==bUseCompletionRtn)                      // have a completion routine?
      {
       KeInitializeEvent(
                         &lclEvent,
                         NotificationEvent,
                         FALSE                        // initialize to not-signalled.
                        );
                         
       IoCopyCurrentIrpStackLocationToNext(pIrp);     // copy parameters from current stack position to next-lower position.

       IoSetCompletionRoutine(
                              pIrp,
                              (PIO_COMPLETION_ROUTINE)JAf2DriverCompletionRtn2,
                              &lclEvent,              // object to be signalled.
                              TRUE,                   // invoke on success.
                              TRUE,                   // invoke on error.
                              TRUE                    // invoke on cancel.
                             );
                              
       InterlockedIncrement(&pDevExt->ulRtn2Needed1); // bump count.
      }
    else
      {
       IoSkipCurrentIrpStackLocation(pIrp);           // copy parameters from current stack position to next-lower position.
                                                      //   No completion routine.
      }

    status = IoCallDriver(                            // call next-lower driver.
                          pDevExt->pNextLowerDevObj,
                          pIrp
                         );

    if (TRUE==bUseCompletionRtn)                      // have a completion routine?
      {
       DbgPrint(("%s DispatchAny:  Status from IoCallDriver = 0x%08x\n", pMstrDevExt->DriverName.Buffer, status));

       DbgPrint(("%s DispatchAny:  Irp at 0x%08x, pIrp->IoStatus.Status = 0x%08x\n",
                 pMstrDevExt->DriverName.Buffer, pIrp, pIrp->IoStatus.Status));

//     if (pIrp->PendingReturned)
//       {
//        DbgPrint(("%s DispatchAny: Irp pending-returned\n", pMstrDevExt->DriverName.Buffer));
//       }
                                                                                                           
//     DbgPrint(("%s DispatchAny:     About to wait\n", pMstrDevExt->DriverName.Buffer));

       bCompleteReq = TRUE;                           // testing.

       InterlockedIncrement(&pDevExt->ulRtn2Needed2); // bump count.

       KeWaitForSingleObject(                         // waiting for JAf2DriverCompletionRtn2.
                             &lclEvent,
                             Executive,               // routine waiting on its own behalf.
                             KernelMode,              // ensure stack (where event object is) won't be paged out.
                             FALSE,                   // not alertable.
                             NULL                     // no timeout value.
                            );

//     DbgPrint(("%s DispatchAny:     Done with wait\n", pMstrDevExt->DriverName.Buffer));

       status = pIrp->IoStatus.Status;                // from \src\general\toaster\func\toaster.c,
                                                      //      routine ToasterSendIrpSynchronously
                                                      // also \src\storage\fdc\flpydisk\floppy.c,
                                                      //      routine FloppyPnp.
       IoCompleteRequest(pIrp, IO_NO_INCREMENT);      // complete request.
      }
   }

done:

 if (TRUE==bHaveRemoveLock)                           // have removal lock?
   IoReleaseRemoveLock(
                       &pDevExt->RemoveLock,
                       NULL                           // didn't supply caller's tag.
                      );

 DbgPrint(("%s DispatchAny:    Final status = 0x%08x\n", pMstrDevExt->DriverName.Buffer, status));
 DbgPrint(("                           Irp status   = 0x%08x\n", pIrp->IoStatus.Status));

 return status;
}

/**************************************************************************************************/
/*                                                                                                */
/* Adapted very closely from \src\general\toaster\toastmon\toastmon.c in DDK, routine             */
/* ToastMon_DispatchSystemControl.                                                                */
/*                                                                                                */
/* Needs more work, since checked kernel says not properly registered.                            */
/*                                                                                                */
/**************************************************************************************************/
NTSTATUS
JAf2DriverDispatchWMI(
                      IN PDEVICE_OBJECT pDeviceObject,
                      IN PIRP           pIrp
                     )
{
 NTSTATUS                status;
 pJAf2DeviceExtension    pDevExt =                    // address of device object extension.
                           (pJAf2DeviceExtension)pDeviceObject->DeviceExtension,
                         pMstrDevExt = (pJAf2DeviceExtension)pDevExt->pMstrDevObj->DeviceExtension;
 BOOLEAN                 bCompleteReq = FALSE,
                         bUseCompletionRtn = FALSE,
                         bHaveRemoveLock = FALSE;

 DbgPrint(("%s DispatchWMI:    Device object at 0x%08x\n", pMstrDevExt->DriverName.Buffer, pDeviceObject));

 // Set the request up as successful
  
 status = IoAcquireRemoveLock(
                              &pDevExt->RemoveLock,
                              NULL                    // don't supply caller's tag.                              
                             );

 if (!NT_SUCCESS(status))                             // no good?
   {
    DbgPrint(("%s DispatchWMI:  Failed to get RemoveLock!\n", pMstrDevExt->DriverName.Buffer));

    pIrp->IoStatus.Status = status;              
    bCompleteReq = TRUE;                              // so request to be completed.
       
    goto done; 
   }

 bHaveRemoveLock = TRUE;

 IoSkipCurrentIrpStackLocation(pIrp);                 // copy parameters from current stack position to next-lower position.
                                                      //   No completion routine.

 status = IoCallDriver(                               // call next-lower driver.
                       pDevExt->pNextLowerDevObj,
                       pIrp
                      );
                       
done:

 if (TRUE==bCompleteReq)                              // I/O request to be completed?
   {
    IoCompleteRequest(pIrp, IO_NO_INCREMENT);         // complete request.
   }

 if (TRUE==bHaveRemoveLock)                          // have removal lock?
   IoReleaseRemoveLock(
                       &pDevExt->RemoveLock,
                       NULL                           // didn't supply caller's tag.
                      );

 DbgPrint(("%s DispatchWMI:    Final status = 0x%08x\n", pMstrDevExt->DriverName.Buffer, status));
 DbgPrint(("                           Irp status   = 0x%08x\n", pIrp->IoStatus.Status));

 return status;
}

/**************************************************************************************************/
/*                                                                                                */
/* JAf2DriverUnload                                                                               */
/*                                                                                                */
/* Clean up.                                                                                      */
/*                                                                                                */
/**************************************************************************************************/
VOID
JAf2DriverUnload(
                 IN PDRIVER_OBJECT pDriverObject
                )
{
 DbgPrint((JAf2DrvNm " JAf2DriverUnload:  Unloading\n"));
}

/**************************************************************************************************/
/*                                                                                                */
/* JAf2DriverCompletionRtn                                                                        */
/*                                                                                                */
/* Simply propagates pending status.                                                              */
/*                                                                                                */
/**************************************************************************************************/
NTSTATUS                                                                                                   
JAf2DriverCompletionRtn(                                                                                   
                        IN PDEVICE_OBJECT pDeviceObject,                                                   
                        IN PIRP           pIrp,                                                            
                        IN PVOID          pCtx                                                             
                       )                                                                                   
{                                                                                                          
 NTSTATUS                status = STATUS_PENDING;                                                          
 pJAf2DeviceExtension    pDevExt =                    // address of device object extension.
                           (pJAf2DeviceExtension)pDeviceObject->DeviceExtension,
                         pMstrDevExt = (pJAf2DeviceExtension)pDevExt->pMstrDevObj->DeviceExtension;
                                                                                                           
 DbgPrint(("%s CompletionRtn:  Entered\n", pMstrDevExt->DriverName.Buffer));                                             
                                                                                                           
 if (pIrp->PendingReturned)                           // per Oney, p 194, and per DDK under IoMarkIrpPending().
   IoMarkIrpPending(pIrp);
                                                                                                           
 return status;                                                                                            
}                                                                                                         

/**************************************************************************************************/
/*                                                                                                */
/* JAf2DriverCompletionRtn2                                                                       */
/*                                                                                                */
/* Signals whatever's waiting.                                                                    */
/*                                                                                                */
/**************************************************************************************************/
NTSTATUS                                                                                                   
JAf2DriverCompletionRtn2(                                                                                   
                         IN PDEVICE_OBJECT pDeviceObject,                                                   
                         IN PIRP           pIrp,                                                            
                         IN PVOID          pCtx                                                             
                        )                                                                                   
{                                                                                                          
 NTSTATUS                status = STATUS_MORE_PROCESSING_REQUIRED;                                                          
 pJAf2DeviceExtension    pDevExt =                    // address of device object extension.
                           (pJAf2DeviceExtension)pDeviceObject->DeviceExtension,
                         pMstrDevExt = (pJAf2DeviceExtension)pDevExt->pMstrDevObj->DeviceExtension;
  
 DbgPrint(("%s CompletionRtn2:  Entered\n", pMstrDevExt->DriverName.Buffer));                                             
                                                                                                           
 if (pIrp->PendingReturned)                           
   {
//  DbgPrint(("%s CompletionRtn2:  Marking irp at 0x%08x\n", pMstrDevExt->DriverName.Buffer, pIrp));                                             
     
    IoMarkIrpPending(pIrp);
   }
                                                                                                           
 InterlockedIncrement(&pDevExt->ulRtn2Entered);       // bump count.
           
 KeSetEvent(                                                                                               
            pCtx,
            IO_NO_INCREMENT,
            FALSE                                     // running (presumably) at higher than PASSIVE_LEVEL.
           );                                                                                              
                                                                                                           
 return status;                                                                                            
}                                                                                                         

/**************************************************************************************************/
/*                                                                                                */
/**************************************************************************************************/
int
JAf2DriverTechnique(
                    IN PDEVICE_OBJECT pDeviceObject,
                    IN PIRP           pIrp,
                    IN PVOID          pInBuffer,
                    IN ULONG          lenInBuffer,
                    IN PVOID          pOutBuffer,
                    IN ULONG          lenOutBuffer
                   )
{
 UNICODE_STRING                  DriverRegistryNm,
                                 WkUniStr,
                                 FullDriverPath;
 ULONG                           rc = TRUE;
 ULONG                     const constOne = 1,
                                 constThree = 3;
 NTSTATUS                        status;
 OBJECT_ATTRIBUTES               KeyObj;
 #define INVALID_HANDLE_VALUE    -1

 goto done;

done:

 return rc;
}

