//----------------------------------------------------------------------
//
// FilterDrv Installation Program
//
// Adapted from install\ctrl2cap.c, part of Russinovich's ctrl2cap package 
// at www.sysinternals.com.
//
// To change:
//
//   1) DriveName
//   2) SysFileSource
//   3) FilterDrv_DRIVER_KEY
//   4) FilterDrv_FILTER_KEY
//
//----------------------------------------------------------------------
#include <windows.h>
#include <stdio.h>

#define JAPgmVer  "1.02"
#define JAPgmName "LoadFilter"

 #define DriverName "JAf2Driver"
// #define DriverName "devupper"
// #define DriverName "filter"

 char SysFileSource[] = "c:\\ja.pgm\\Driver\\libchk\\i386\\" DriverName ".sys";
// char SysFileSource[] = "d:\\Win2KDDK\\src\\general\\toaster\\filter\\devupper\\objchk\\i386\\" DriverName ".sys";
// char SysFileSource[] = "d:\\Win2KDDK\\src\\wdm\\general\\" DriverName "\\objchk\\i386\\" DriverName ".sys";

//
// Registry keys we access
//

 #define FilterDrv_DRIVER_KEY   "System\\CurrentControlSet\\Services\\JAf2Driver"
// #define FilterDrv_DRIVER_KEY   "System\\CurrentControlSet\\Services\\devupper" 
// #define FilterDrv_DRIVER_KEY   "System\\CurrentControlSet\\Services\\Filter"

// Next is for CDROM.
 
 #define FilterDrv_FILTER_KEY   "System\\CurrentControlSet\\Control\\Class\\{4D36E965-E325-11CE-BFC1-08002BE10318}"
 
// Next is for floppy.
 
// #define FilterDrv_FILTER_KEY   "System\\CurrentControlSet\\Control\\Class\\{4D36E980-E325-11CE-BFC1-08002BE10318}"

//----------------------------------------------------------------------
//
// SearchMultiSz
//
// Finds the specified string in a multiSz and returns the address of
// the found string (or NULL if not found).
//
//----------------------------------------------------------------------
char *
SearchMultiSz( 
  char *Value,                                        // string to search.  
  DWORD ValueLength,                                  // length of string to search. 
  char *String                                        // string to search for. 
  )
{
  DWORD len;

  if( ValueLength < strlen( String )) return NULL;

  len = ValueLength - strlen(String);
  do {
      if( !stricmp( &Value[len], String ) &&
                 (!len || !Value[len-1])) {

                 return &Value[len];
             }
  } while( len-- );
  return NULL;
}


//----------------------------------------------------------------------
//
// PrintError
//
// Presents an error as an easy to read message.
//
//----------------------------------------------------------------------
void
PrintError(
   DWORD Code
   )
{
   CHAR                                     buffer[80] ;
   DWORD                                    count ;

   //
   // Translate the Win32 error code into a useful message.
   //
   count = FormatMessage (FORMAT_MESSAGE_FROM_SYSTEM,
                          NULL,
                          Code,
                          0,
                          buffer,
                          sizeof (buffer),
                          NULL) ;

   //
   // Make sure that the message could be translated.
   //
   if (count == 0) {

      printf("\nError could not be translated.\n Code: %d\n", Code) ;
      return;
   }
   else {

      //
      // Display the translated error.
      //
      printf("%s\n", buffer) ;
      return;
   }
}


//----------------------------------------------------------------------
// 
// FilterDrvUninstall
//
// Uninstalls the filter driver.
//
//----------------------------------------------------------------------
void 
FilterDrvUninstall(
    BOOLEAN Silent,
    DWORD Version, 
    char *SystemDirectory 
    )
{
    HKEY  hKey;
    DWORD error, type, length;
  char  filters[MAX_PATH], *ptr;

#if 0                                                 // Leave this part out.  ja.haw.  08.05.01.

    //
    // Delete the driver file.
    //
    if( !DeleteFile( SystemDirectory ) && !Silent) {

        printf("Error deleting \\system32\\drivers\\" DriverName ".sys:\n");
        PrintError( GetLastError());
    }

    //
    // Delete the driver Registry key.
    //
  RegDeleteKey( HKEY_LOCAL_MACHINE, FilterDrv_DRIVER_KEY"\\Security" );
  RegDeleteKey( HKEY_LOCAL_MACHINE, FilterDrv_DRIVER_KEY"\\Enum" );
  if(( error = RegDeleteKey( HKEY_LOCAL_MACHINE, FilterDrv_DRIVER_KEY )) !=
    ERROR_SUCCESS && !Silent ) {

    printf("Error deleting " DriverName " driver key:\n");
    PrintError( error );
  }

#endif

  //
  // Delete filter from cdrom filter value
  //
  if( Version > 4 ) {

    if( (error = RegOpenKey( HKEY_LOCAL_MACHINE,
            FilterDrv_FILTER_KEY,
            &hKey )) != ERROR_SUCCESS ) {

      if( !Silent ) {
        printf("Could not open " DriverName " filter Registry key:\n");
        PrintError( error );
      }
      return;     
    }

    length = sizeof( filters );
    if( (error = RegQueryValueEx( hKey, "UpperFilters", 0, &type, filters, &length )) !=
        ERROR_SUCCESS ) {

      if( !Silent ) {

        error = GetLastError();
        printf("Could not query " DriverName " filter Registry key:\n");
        PrintError( error );
        RegCloseKey( hKey );
      }
      return;
    } 
    if( ptr = SearchMultiSz( filters, length, DriverName )) {

      //
      // Zap it.
      //
      memcpy( ptr, ptr + strlen(DriverName)+1,
          length - (ptr-filters) - strlen(DriverName) -1 );
      length -= strlen(DriverName)+1;

      if( (error = RegSetValueEx( hKey, "UpperFilters", 0, type,
            filters, length )) != ERROR_SUCCESS ) {

        if( !Silent ) {

          printf("Could not reset " DriverName " filter Registry key:\n");
          PrintError( error );
        }
      }
    }
    RegCloseKey( hKey );
  }
  printf("Filter uninstalled.  You must reboot for this to take effect.\n\n");
}


//----------------------------------------------------------------------
//
// FilterDrvInstall
//
// Installs the filter driver.
//
//----------------------------------------------------------------------
void 
FilterDrvInstall(
    DWORD Version, 
    char *SystemDirectory 
    )
{
    HKEY  hKey;
    DWORD error, value, type, length;
  char  filters[MAX_PATH];

    //
    // First, copy the driver file to the system directory.
    //
   if( Version > 4 ) {

        if( !CopyFile( SysFileSource, SystemDirectory, FALSE )) {

            printf("Could not copy " DriverName ".sys to \\system32\\drivers.:\n");
            PrintError( GetLastError());
            exit(1);
        }
    }          
   else
     {
      printf("Level of OS is before Win2000, quitting.\n");
      exit(1);
     }

    //
    // Create the driver Registry key.
    //
    if( (error = RegCreateKey( HKEY_LOCAL_MACHINE,
                       FilterDrv_DRIVER_KEY,
                       &hKey )) != ERROR_SUCCESS ) {

    error = GetLastError();
        PrintError( error );
        FilterDrvUninstall( TRUE, Version, SystemDirectory );
        exit(1);
    }

    //
    // Add the appropriate values.
    //
    value = 1;
    RegSetValueEx( hKey, "Type", 0, REG_DWORD, (PCHAR) &value, sizeof(value));
    value = 1;
    RegSetValueEx( hKey, "ErrorControl", 0, REG_DWORD, (PCHAR) &value, sizeof(value));
    
    //
    // On Win2K, add the driver as a cdrom filter.
    //
    if( Version > 4 ) {

    value = 3;                                        // ja/haw.  06.05.01.
    RegSetValueEx( hKey, "Start", 0, REG_DWORD, (PCHAR) &value, sizeof(value));
    value = 1;
    RegCloseKey( hKey );
    if( (error = RegOpenKey( HKEY_LOCAL_MACHINE,
            FilterDrv_FILTER_KEY,
            &hKey )) != ERROR_SUCCESS ) {

      printf("Could not create " DriverName " filter Registry key:\n");
      goto error;
    }

    length = sizeof( filters );
    if( (error = RegQueryValueEx( hKey, "UpperFilters", 0, &type, filters, &length )) !=
        ERROR_SUCCESS ) {

      if (ERROR_FILE_NOT_FOUND!=error)                // something other than 'not found'?
        {
         error = GetLastError();
         printf("Could not open " DriverName " filter Registry key:  Error = %8xx\n", error);
         RegCloseKey( hKey );
         goto error;
        }

      length = 1;                                     // build empty multiple null string array;
      filters[length-1] = '\0';
      type = REG_MULTI_SZ;
    }

    // 
    // Append filter to the end and update the key value.
    //
                                                              
    if( !SearchMultiSz( filters, length, DriverName ))  
      {

      strcpy( &filters[length-1], DriverName);
      length = length + strlen(DriverName);
      filters[ length ] = 0;

      if( (error = RegSetValueEx( hKey, "UpperFilters", 0, type,
        filters, length + 1 )) != ERROR_SUCCESS ) {

        error = GetLastError();
        printf("Could not set " DriverName " filter Registry key:\n");
        RegCloseKey( hKey );
        goto error;
      }
    }
    RegCloseKey( hKey );
    } else {

    value = 3;                                        // ja/haw.  06.05.01.
    RegSetValueEx( hKey, "Start", 0, REG_DWORD, (PCHAR) &value, sizeof(value));
    value = 1;
    RegSetValueEx( hKey, "Group", 0, REG_SZ, "KeyboardClass", strlen("KeyboardClass")+1);
    RegCloseKey( hKey );
  }

  printf("Filter successfully installed.  You must reboot for it to take effect.\n\n");
  return;

error:
  //
  // Cleanup path when we have to uninstall stuff on the way out.
  //
  PrintError( error );
  FilterDrvUninstall( TRUE, Version, SystemDirectory );
  exit(1);
}


//----------------------------------------------------------------------
//
// Usage
// 
// Tell the user the proper arguments.
//
//----------------------------------------------------------------------
void 
Usage( 
    char *name 
    )
{
    printf("usage: %s [/install | /uninstall]\n\n", name );
    exit(1);
}


//----------------------------------------------------------------------
//
// Main
// 
// Just prints banner and calls the uninstall or install function.
//
//----------------------------------------------------------------------
void 
main( 
    int argc, 
    char *argv[] 
    )
{
    DWORD   version;
    CHAR    systemDirectory[MAX_PATH] ;

    printf(JAPgmName " " JAPgmVer " started.\n");

    if( argc != 2 ||
        (stricmp( argv[1], "/install") && stricmp( argv[1], "/uninstall") )) {
        
        Usage( argv[0] );
    }

    //
    // Get NT version
    //
    version = GetVersion();
     
//  if( version >= 0x80000000 ) {
//
//      printf("This version of FilterDrv works on NT 4 and Win2K.\n"
//             "Visit http://www.sysinternals.com for a version that works on Windows 9x\n\n");
//      return;
//  }

    //
    // Get the appropriate directory
    //
    GetSystemDirectory( systemDirectory, sizeof (systemDirectory)) ;
    strcat( systemDirectory, "\\drivers\\" DriverName ".sys") ;

    //
    // Do the install or uninstall routine.
    //
    if( !stricmp( argv[1], "/install" )) {

        FilterDrvInstall( version & 0xFF, systemDirectory );

    } else {
        
        FilterDrvUninstall( FALSE, version & 0xFF, systemDirectory );
    }
}
